#include "finauto.h"
     
bool FiniteAutomaton::ToStatesForEvent(
   int i_from, char i_oninput, CIntSet& o_toset)
{
  return pFaQuery->ToStatesForEvent(i_from,
                                    i_oninput,
                                    o_toset);
}
     
int FiniteAutomaton::GetAcceptState() 
{
  return pFaQuery->GetAcceptState();
}
     
FiniteAutomaton::FiniteAutomaton(CFAQuery& iquery) 
   :pFaQuery(&iquery)
{
}
     
///////////////////////////////////////////////// 
// AutomatonState
// State interator for FiniteAutomaton
     
// only to be used by FiniteAutomaton::Create... Iterator methods
AutomatonState::AutomatonState(
   FiniteAutomaton *a_parent, int startState)
   : parent(a_parent)
{
   // Initialise state set with start state 
   currentState.Append(startState);
   // ... and traverse all empty states from here 
   TraverseEvents(currentState, currentState, emptyEvent);
}
     
AutomatonState::~AutomatonState()
{ }
     
AutomatonState::AutomatonState(const AutomatonState &src)
{
   *this = src;
}
     
AutomatonState &
AutomatonState::operator=(const AutomatonState &src)
{
   if (this != &src)
   {
      parent = src.parent;
      currentState = src.currentState;
   }
     
   return(*this);
}
     
bool AutomatonState::IsAccept()
{
   return(currentState.Find(parent->GetAcceptState())
                           != -1);
}
     
bool AutomatonState::IsValid()
{
   return(currentState.GetSize() > 0);
}
     
// Accepts an event and uses the state machine to move onto the 
// next state. Returns true if the state is non-terminal, ie 
// not accept, and valid. Note: we treat accept as being a 
// terminal state here for the sake of consistency with a 
// notional deterministic automaton interface, and also because 
// in many cases finding an accept state is sufficient reason to 
// halt further processing 
bool AutomatonState::NextState(char event)
{
   PerformTransitionSet(event);
   return(IsValid() && !IsAccept());
}
     
// Method to do the non-deterministic transitions. The algorithm is 
// to follow all transitions from the current set of states which 
// correspond to the passed event until there are no more 
// transitions to follow for the current event. Empty event 
// transitions are treated as always valid. Note: if this method is 
// called with an accept position in the currentState set, that 
// position may be lost.
void AutomatonState::PerformTransitionSet(char event)
{
   // Reject if currentState is not valid 
   if (!IsValid())
   {
      return;
   }
     
   // In a deterministic automaton, we would not perform any further 
   // traversal if we had reached an accept state. However, a non- 
   // deterministic automaton is not necessarily in a terminal state 
   // if the accept state has been found
     
   // Storage for new state set
   CIntSet nextState;
     
   TraverseEvents(currentState, nextState, event);
   currentState = nextState;
     
   //cout << "pre-closure - ";
   //Dump(cout);
     
   TraverseEvents(currentState, currentState, emptyEvent);
}
     
// Helper function to handle traversal of available transitions for 
// given event from positions in source set. Adds states traversed 
// to onto target set. If source and target are the same, the 
// effect will be to traverse all adjacent transitions for the 
// passed event (this is useful for closure of empty transitions).
void AutomatonState::TraverseEvents(
   CIntSet &source, CIntSet &target, char event)
{
   for (int pos = 0; pos < source.GetSize();
            pos++)
   {
   CIntSet toset;
   int from = source[pos];
   parent->ToStatesForEvent(from, event, toset);
      target.Append(toset);
   }
}
